import sys
from distutils.version import LooseVersion

from qai.meta import pypi_versions


if __name__ == "__main__":
    local_version = sys.argv[1]

    pypi_versions = pypi_versions.get_pypi_versions()
    print(f"found the following versions on pypi: {pypi_versions}")
    pypi_max_version = pypi_versions[-1]

    if LooseVersion(local_version) > LooseVersion(pypi_max_version):
        print(
            f"found local version {local_version} and highest version on PyPi {pypi_max_version}"
        )
        exit(0)
    else:
        print(
            "\nThe local VERSION is not greater than the largest already published version on PyPi!"
            " Please increment the VERSION before trying to release.\n"
        )
        exit(1)



from qai.cleantext import Cleantext

def scan_tokens(token):
    return str(token.dep_) in ["auxpass", "csubjpass", "nsubjpass"]

def is_passive_voice(phrase):
    sent = Cleantext(use_spacy=True).segment_sentences(phrase)
    passive = False
    for token in sent[0]:
        if scan_tokens(token):
            passive = True
    
    return passive



from typing import List, Optional, Tuple


def get_meta_value(
    category: str, meta: List[object]
) -> Tuple[bool, Optional[int], List]:
    """
    The business logic is:
        * read it from chain.meta
            * it is Optionaal[JsValue]
            * enabled will never be 0, per Muayad, so enabled is always True,
              we should deprecate it
    More on enabled: if we get a message, the service is enabled,
    hence the first return value is now always True
    """
    enabled, value, sub_groups = True, None, []
    if not meta:
        pass
        print("Empty meta field. Attempting to process anyway; expect errors")
    else:
        if len(meta):
            try:
                value = int(meta["value"])
            except (KeyError, ValueError):
                # apparently for on/off services, this is
                # now the service name
                value = None

            try:
                sub_groups = meta["subGroup"]
            except KeyError:
                sub_groups = []

    return True, value, sub_groups



import uuid


def make_issue(
    word: str,
    start: int,
    end: int,
    issue_type: str,
    score,
    description="",
    suggestions=[],
    subCategory="",
):
    """
    reference:
    https://qordoba.atlassian.net/wiki/spaces/HOME/pages/840368141/Content-AI

    Note on duplicate value issueId / id: the platform switched this key at some point
    But I don't trust being able to delete things, so I will just duplicate it, and one day during a refactor
    we can remove it :grimace:
    """
    _uuid = str(uuid.uuid4())
    if not description:
        description = f'"{word}" is correlated with {issue_type} language.'
    issue = {
        "issueId": _uuid,
        "id": _uuid,
        "issueType": issue_type,
        "from": start,
        "until": end,
        "score": score,
        "suggestions": suggestions,
        "description": description,
    }

    if subCategory:
        issue["meta"] = {"subCategory": subCategory}

    return issue



import re

from typing import List

from spacy.tokenizer import Tokenizer


class Validator:
    html_pattern = re.compile("<.*?>")
    non_letter_pattern = re.compile("[^a-zA-Z'\-\s]")
    acceptable_fraction_of_ignorable_tokens = 0.5

    def __init__(
        self, nlp_obj, sentence_token_limit=1024, ignore_html=True,
    ):
        self.sentence_token_limit = sentence_token_limit
        self.ignore_html = ignore_html
        self.nlp = self.confirm_nlp(nlp_obj)
        self.tokenizer = Tokenizer(self.nlp.vocab)

    def confirm_nlp(self, nlp_obj):
        if not nlp_obj:
            # make the minimal nlp instance needed for SBD
            from spacy.lang.en import English

            nlp = English()  # just the language with no model
            sentencizer = nlp.create_pipe("sentencizer")
            nlp.add_pipe(sentencizer, first=True)
        elif "parser" in nlp_obj.pipe_names:
            # spaCy uses the parser to do SBD
            # so if parser is in the pipe, SBD will happen
            nlp = nlp_obj
        else:
            # we actually need to add SBD to the spacy instance
            # Sentencizer has to be first in the pipeline or you get the error:
            # "ValueError: [E043] Refusing to write to token.sent_start if its document is parsed,
            # because this may cause inconsistent state.""
            nlp = nlp_obj.add_pipe(nlp_obj.create_pipe("sentencizer"), first=True)
        return nlp

    def has_html(self, segment: str):
        return self.html_pattern.match(segment) != None

    def has_non_letter(self, segment: str):
        return self.non_letter_pattern.match(segment) != None

    def is_empty(self, segment: str):
        return len(segment.strip()) == 0

    def has_sentence_over_length_limit(self, segment: str):
        too_long = False
        for sent in self.nlp(segment).sents:
            if len(sent) > self.sentence_token_limit:
                too_long = True
        return too_long

    def is_unnacceptable(self, segment: str):
        is_garbage = False

        if self.ignore_html and self.has_html(segment) is True:
            is_garbage = True

        if self.is_empty(segment):
            is_garbage = True

        if self.has_sentence_over_length_limit(segment):
            is_garbage = True

        return is_garbage

    def _get_ignored_tokens(self, segment: str) -> List[str]:
        ignored_tokens = []
        for token in self.tokenizer(segment):
            if token.like_url:
                ignored_tokens.append(token.text)
            elif "=" in token.text:
                ignored_tokens.append(token.text)
            elif self.has_non_letter(token.text):
                ignored_tokens.append(token.text)
        return ignored_tokens

    def has_too_many_ignore_tokens(self, segment: str) -> bool:
        token_length = len(self.nlp(segment.strip()))
        ignored_tokens = self._get_ignored_tokens(segment)
        ignored_length = len(ignored_tokens)
        if token_length == 0:
            return False
        else:
            return (
                ignored_length / token_length
                > self.acceptable_fraction_of_ignorable_tokens
            )

    def __call__(self, segment: str):
        is_valid = True
        if self.is_unnacceptable(segment):
            is_valid = False
        if self.has_too_many_ignore_tokens(segment):
            is_valid = False
        return is_valid



import pprint
import os
import sys

from difflib import Differ
from typing import Tuple

from qai import __version__


pp = pprint.PrettyPrinter(indent=2)


def find_files(dir_abspath) -> Tuple[str, str, str]:
    # find application
    app = os.path.join(dir_abspath, "app.py")
    application = os.path.join(dir_abspath, "application.py")
    if not (os.path.exists(app) or os.path.exists(application)):
        raise FileNotFoundError(
            f"there is no app.py or application.py in the provided directory, {dir_abspath}"
        )
    app_path = app if os.path.exists(app) else application

    docker_path = os.path.join(dir_abspath, "Dockerfile")
    if not (os.path.exists(docker_path)):
        raise FileNotFoundError(
            f"there is no Dockerfile in the provided directory, {dir_abspath}"
        )

    requirements_path = os.path.join(dir_abspath, "requirements.txt")
    if not (os.path.exists(requirements_path)):
        with open(requirements_path, "w+") as f:
            f.write("")

    return app_path, docker_path, requirements_path


class DockerFile:
    new_base = "qsam/spacy_alpine:37"

    def __init__(self, fp):
        self.name = "Dockerfile"
        self.fp = fp
        with open(fp, "r") as f:
            self.lines = f.readlines()
        self.condition_modifier = [
            (
                lambda i, line: "FROM gcr.io/qordoba-devel/qai:" in line,
                self.from_replace,
            ),
            (lambda i, line: "WORKDIR /qai" in line, self.delete_line),
            (
                lambda i, line: "#" in line
                and "only" in line
                and "deps" in line
                and "qai" in line,
                self.delete_line,
            ),
            (lambda i, line: "pip install ." in line, self.delete_line),
        ]

    def from_replace(self, i, line):
        self.lines[i] = f"FROM {self.new_base}"

    def delete_line(self, i, line):
        self.lines[i] = ""

    def parse_lines(self):
        """
        This script will be run ~6 times.
        So better for the script to be slow, but writting it to be fast

        This method should be idempotent
        """
        for i, line in enumerate(self.lines):
            for condition, modifier in self.condition_modifier:
                if condition(i, line):
                    modifier(i, line)

    def dry_run(self):
        self.parse_lines()
        return "".join(self.lines)

    def update_DESTRUCTIVE(self):
        with open(self.fp, "w+") as f:
            f.write(self.dry_run())


class RequirementsTxt:
    def __init__(self, fp):
        self.name = "requirements.txt"
        self.fp = fp
        with open(self.fp, "r+") as f:
            self.req = f.read()
        self.qai = f"qai>={__version__}"

    def dry_run(self):
        return self.req.rstrip("\n") + f"\n{self.qai}\n"

    def update_DESTRUCTIVE(self):
        with open(self.fp, "w+") as f:
            f.write(self.dry_run())


class ApplicationDotPy:
    def __init__(self, fp):
        self.name = "application.py" if "application" in fp else "app.py"
        self.fp = fp
        with open(self.fp, "r") as f:
            self.lines = f.readlines()

    def add_connect_call_post_qconnect(self, i, line, start_pos=None, depth=0):
        if start_pos is None:
            depth = 1
            Qconnect = "QConnect("
            offset = len(Qconnect)
            start_pos = line.index(Qconnect) + offset
        for i_char, c_char in enumerate(line[start_pos:]):
            if c_char == "(":
                depth += 1
            elif c_char == ")":
                depth -= 1

            if depth == 0:
                print("found closing paren")
                newline = line[: i_char + offset + 1].rstrip() + ".connect()"
                self.lines[i] = newline

        if depth > 0:
            # QConnect(... must coninue on next line
            self.lines[i] = line.rstrip() + self.lines[i + 1].strip()
            del self.lines[i + 1]
            self.add_connect_call_post_qconnect(i, self.lines[i])

        if depth > 5:
            raise RecursionError(
                "Can't automatically replace QConnect with QRest, do it yourself"
            )

    def fix(self):
        for i, line in enumerate(self.lines):
            if "from qai" in line and "import QRest" in line:
                self.lines[i] = "from qai.qconnect.qrest import QRest"
            if "import QConnect" in line:
                self.lines[i] = "from qai.qconnect.qrest import QRest"
            if "QConnect(" in line:
                self.add_connect_call_post_qconnect(i, line)
                self.lines[i] = self.lines[i].replace("QConnect", "QRest")

    def dry_run(self):
        self.fix()
        return "".join(self.lines)

    def update_DESTRUCTIVE(self):
        with open(self.fp, "w") as f:
            f.write(self.dry_run())


def command_line():

    try:
        supplied_dir = sys.argv[1]
    except IndexError:
        supplied_dir = None

    if supplied_dir is None or not os.path.isdir(supplied_dir):
        raise ValueError(
            "Call this script with the path of the directory you want to fix"
            "e.g. `python -m qai.upgrade .` or `python -m qai.upgrade Users/sh/q/repos/service.formality`"
        )

    SERVICE_DIR = os.path.abspath(sys.argv[1])
    app_path, docker_path, requirements_path = find_files(SERVICE_DIR)

    with open(app_path, "r") as af, open(docker_path) as df, open(
        requirements_path
    ) as rf:
        old_files = [af.readlines(), df.readlines(), rf.readlines()]

    fixers = [
        ApplicationDotPy(app_path),
        DockerFile(docker_path),
        RequirementsTxt(requirements_path),
    ]

    outputs = [x.dry_run().splitlines(keepends=True) for x in fixers]

    print("The following changes would be applied:\n\n")

    d = Differ()
    for i in range(len(old_files)):
        print(f"{fixers[i].name}:\n")
        result = list(d.compare(old_files[i], outputs[i]))
        pp.pprint(result)
        print("\n")

    should_fix = input("Apply these changes? y or n")

    if should_fix != "y":
        print("no changes applied.")
        exit(0)

    for x in fixers:
        print(f"\nfixing {x.name}\n")
        x.update_DESTRUCTIVE()
    print("\n\ndone\n\n")




from qai.issues.add_issues import parse


def get_payload_stats(json_data, nlp):
    try:
        _ = json_data.items()
        els = [json_data]
    except AttributeError:
        els = json_data

    n_seg = len(els) if els else 0
    n_sent = 0
    n_word = 0

    try:
        for el in els:
            segment, _, _ = parse(el)
            n_sent += len(list(nlp(segment).sents))
            n_word += len(segment.split())
    except:
        # this will cause escaped exception in sanic logger
        # do nothing
        pass

    return n_seg, n_sent, n_word



import math
import multiprocessing
import os
import time
from pathlib import Path

from sanic import Sanic, response
from sanic.log import logger

from qai.funq import partial
from qai.issues.add_issues import (
    add_issues_format_insensitive,
    add_issues_format_insensitive_batch,
)
from qai.issues.logs import get_payload_stats
from qai.issues.validation import Validator
from qai.qconnect.qremoteconnection import QRemoteConnection
from qai.version import __version__


def get_cpu_quota_within_docker():
    """
    By default, we use mp.cpu_count()
    HOWEVER, if there are cpu limits in certain paths, we assume those are
    from k8s/docker and override with those values
    """
    cpu_cores = multiprocessing.cpu_count()

    cpu_shares = Path("/sys/fs/cgroup/cpu/cpu.shares")

    if cpu_shares.exists():
        with cpu_shares.open("rb") as r:
            request_cpu_shares = int(r.read().strip())
            cpu_cores = (
                math.ceil(request_cpu_shares / 1024)
                if request_cpu_shares > 0
                else multiprocessing.cpu_count()
            )
            print(
                f"CPU request limit: {round(request_cpu_shares / 1024, 2)}, will spin {cpu_cores} worker(s)."
            )
    else:
        print(f"CPU_shares not found, will spin {cpu_cores} worker(s).")
    return cpu_cores


async def post_root(self, request):
    json_data = request.json
    try:
        # if it has items it is a dict
        _ = json_data.items()
        # if it is a dict, make it an array
        els = [json_data]
    except AttributeError:
        # can't call .items(), so it's already an array
        els = json_data
    resp_list = []

    if self.batching:
        resp_list = add_issues_format_insensitive_batch(
            self, els, self.debug, verbose=self.verbose
        )
        return response.json(resp_list)

    for el in els:
        el = add_issues_format_insensitive(
            self, el, self.validator, debug=self.debug, verbose=self.verbose
        )
        resp_list.append(el)
    return response.json(resp_list)


async def get_root(self, request):
    try:
        white_lister = self.white_lister
    except AttributeError:
        white_lister = None
    resp_obj = {
        "service": self.category,
        "status": "up",
        "host": self.host,
        "port": self.port,
        "qai_version": __version__,
        "white_lister": str(white_lister),
        "analyzer": str(self.analyzer),
    }
    return response.json(resp_obj)


async def add_start_time(request):
    """Prepend initial time when this request was served."""
    request.ctx.start_time = time.time()


async def add_perf_stats(self, request, response):
    """
    Prepend initial time when this request was served.
    Log the access long on each request.
    Log num of segments, sentences, words in payload.
    Log sentence and word speed.
    """
    is_get = request.method.lower() == "get"
    is_check = request.path in ["/ready", "/healthy"]
    should_log = not (is_get or is_check)
    if should_log:
        latency = round((time.time() - request.ctx.start_time) * 1000)
        n_seg, n_sent, n_word = get_payload_stats(request.json, self.validator.nlp)
        latency_sec = latency / 1000

        try:
            # log how long it takes to process seg/sent
            seg_latency = round(latency_sec / n_seg, 2)
            sent_latency = round(latency_sec / n_sent, 2)

            # how many words per sec
            word_speed = round(n_word / latency_sec, 2)

        except:
            # markers - logging error
            seg_latency = -1
            sent_latency = -1
            word_speed = -1

        logger.info(
            (
                "{method} {url} {status} {latency}ms {reqbytes}reqBytes "
                "{resbytes}resBytes {n_seg}segs {n_sent}sents {n_word}words "
                "{seg_latency}spSeg {sent_latency}spSent {word_speed}wps"
            ).format(
                method=request.method,
                ip=request.ip,
                url=request.url,
                status=response.status,
                latency=latency,
                reqbytes=len(request.body),
                resbytes=len(response.body),
                n_seg=n_seg,
                n_sent=n_sent,
                n_word=n_word,
                seg_latency=seg_latency,
                sent_latency=sent_latency,
                word_speed=word_speed,
            )
        )


class QRest(QRemoteConnection):
    def __init__(
        self,
        analyzer,
        category="",
        white_lister=None,
        host="0.0.0.0",
        port=5000,
        workers=None,
        config_path=["conf", "config.json"],
        batching=False,
        debug=False,
        verbose=False,
        sentence_token_limit=1024,
        ignore_html=True,
        ignore_inside_quotes=False,
    ):
        self.host = host
        self.port = port

        config_file = os.path.join(os.getcwd(), *config_path)
        super().__init__(analyzer, category, white_lister, config_file)

        if workers is None:
            # get workers from configs
            try:
                self.workers = int(self.configs["WORKER_COUNT"])
                print(f"Found WORKER_COUNT in configs, setting it to {self.workers}")
            except KeyError:
                # WORKER_COUNT not in configs
                print(
                    "WORKER_COUNT not set in code or configs, using as many as CPU cores available"
                )
                self.workers = get_cpu_quota_within_docker()
                print(f"set WORKER_COUNT to {self.workers}")
                # Don't catch ValueError, as this happens when we can't coerce WORKER_COUNT to an int
                # in that case we should die and yell at the user..

        elif isinstance(workers, int):
            self.workers = workers
        else:
            raise ValueError(
                (
                    "QRest instantiated with invalid workers value. "
                    "Workers should be set via config, set to an integer value, or not set"
                )
            )
        self.batching = batching

        if self.batching == True and self.workers > 1:
            print("batching set to True - MUST have only 1 worker")
            old_worker_count = self.workers
            self.workers = 1
            print(f"reset workers from {old_worker_count} to {self.workers}")

        self.debug = debug
        self.verbose = verbose
        self.sentence_token_limit = sentence_token_limit
        self.ignore_html = ignore_html
        self.ignore_inside_quotes = ignore_inside_quotes
        nlp_obj = getattr(analyzer, "nlp", None)
        self.validator = Validator(
            nlp_obj=nlp_obj,
            sentence_token_limit=sentence_token_limit,
            ignore_html=ignore_html,
        )

        self.app = Sanic(__name__)

        # Add the request and response middleware
        self.app.request_middleware.append(add_start_time)
        unary_add_perf_stats = partial(add_perf_stats, self)
        self.app.response_middleware.append(unary_add_perf_stats)

        # Add routes
        # This works, but it is better to make these after instantiating the class
        # that way you don't need partial application... see README
        unary_post_root = partial(post_root, self)
        unary_get_root = partial(get_root, self)
        self.app.add_route(unary_post_root, "/", methods=["POST"])
        self.app.add_route(unary_get_root, "/", methods=["GET"])
        self.app.add_route(unary_get_root, "/ready", methods=["GET"])
        self.app.add_route(unary_get_root, "/healthy", methods=["GET"])

    def get_future(self):
        """
        Return a Future (Promise in JS land) that can be put on an event loop
        """
        return self.app.create_server(host=self.host, port=self.port)

    def connect(self):
        """
        Doesn't return, makes a blocking connection
        Only use if you are ONLY using REST
        This is a more robust REST server than get_future makes
        see:
        https://sanic.readthedocs.io/en/latest/sanic/deploying.html#asynchronous-support
        """
        if self.workers > 1:
            print(
                f"grpc, and hence AutoML, only support 1 worker, "
                + f"but you are using {self.workers}"
            )
        return self.app.run(
            host=self.host, access_log=False, port=self.port, workers=self.workers
        )



import traceback
from typing import List, Optional, Tuple

from qai.issues.issues import make_issue

try:
    from spacy.tokens.span import Span
except ModuleNotFoundError as e:
    print(
        "\n\nQAI v2.0.0 and above assume spaCy is installed, but do not list it as a requirement"
        "If you want to use SpacyFactor, please install spaCy >=2.1.0\n\n"
    )
    raise e


class SpacyFactor(object):
    def __init__(self, issue_type: str, description: Optional[str] = None):
        self.issue_type = issue_type
        if description:
            self.description = description

    @staticmethod
    def match_case(span_text, suggestions):
        """
        suggestions respect original casing
        ex. purchase -> buy (lower)
        ex. PURCHASE -> BUY (upper)
        ex. Purchase -> Buy (title)
        """

        try:
            if len(span_text.strip()):
                words = span_text.split(" ")
                if words[0].islower():
                    # todo: exclude I and capitalized proper nouns
                    suggestions = [s.lower() for s in suggestions]

                # len condition: exclude capitalization caused by "I"
                elif len(words[0]) > 1 and words[0].isupper():
                    suggestions = [s.upper() for s in suggestions]

                # title case - first letter check & conversion due to contractions
                elif len(words[0]) > 1 and words[0][0].isupper():
                    for i, s in enumerate(suggestions):
                        if len(s.strip()):
                            s_words = s.split(" ")
                            suggestions[i] = " ".join(
                                [s_words[0][0].upper() + s_words[0][1:]] + s_words[1:]
                            ).strip()
        except:
            suggestions_log = ",".join(suggestions)
            print(traceback.format_exc())
            print(f"Case matching failed for span: {span_text} and {suggestions_log}")

        return suggestions

    def __call__(
        self, spacy_span: Span, meta_subcategory=None, describer=None, cased=True
    ):
        """
        convenient wrapper around make_issue if you are using spaCy

        usage example:

        ```python
        from spacy.tokens import Span
        from app.factor import SpacyFactor


        SOV = SpacyFactor(
            "subject_object_verb_spacing",
            "Keep the subject, verb, and object of a sentence close together to help the reader understand the sentence."
        )

        Span.set_extension("score", default=0)
        Span.set_extension("suggestions", default=[])

        doc = nlp("Holders of the Class A and Class B-1 certificates will be entitled to receive on each Payment Date, to the extent monies are available therefor (but not more than the Class A Certificate Balance or Class B-1 Certificate Balance then outstanding), a distribution.")
        score = analyze(doc)
        if score is not None:
            span = Span(doc, 0, len(doc))  # or whichever TOKENS are the issue (don't have to worry about character indexes)
            span._.score = score
            span._.suggestions = get_suggestions(doc)
            issues = SOV(span)
        ```
        """
        text, start, end = spacy_span.text, spacy_span.start_char, spacy_span.end_char
        score = spacy_span._.score if spacy_span.has_extension("score") else 0
        suggestions = (
            spacy_span._.suggestions if spacy_span.has_extension("suggestions") else []
        )

        if cased:
            suggestions = self.match_case(spacy_span.text, suggestions)

        if describer:
            description = describer(spacy_span)
        else:
            description = self.description

        if meta_subcategory:
            subcategory = meta_subcategory
        else:
            subcategory = ""

        return make_issue(
            text,
            start,
            end,
            issue_type=self.issue_type,
            score=score,
            description=description,
            suggestions=suggestions,
            subCategory=subcategory,
        )



import re
from typing import List

from bs4 import BeautifulSoup


class Segment:
    """
    start: Segment start, calculated from document start
    end: Segment end, calculated from document start
    tag: Segment tag name e.g. span
    html: Segment html
    text: Segment plain text
    start_text: Segment text start, calculated from segment start i.e. return 6 for <span>hello</span>
    end_text: Segment text end, calculated from segment start i.e. return 11 for <span>hello</span>
    attrs: dict containing element html attributes and their values
    """
    start: int
    end: int
    tag: str
    html: str
    text: str
    start_text: int
    end_text: int
    attrs: dict

    def __init__(self, start: int, end: int, tag: str, html: str, text: str = None, start_text: int = None,
                 end_text: int = None, attrs: dict = None):
        self.start = start
        self.end = end
        self.tag = tag
        self.html = html
        if text:
            self.text = text
            self.start_text = start_text
            self.end_text = end_text
        self.attrs = attrs if attrs is not None else {}

    def get_segment_id(self):
        if self.attrs and 'segment_id' in self.attrs:
            return self.attrs['segment_id']
        return None

    def __str__(self) -> str:
        if self.html:
            return self.html
        return super().__str__()


class Paragraph:
    """
    start: Paragraph start, calculated from document start
    end: Paragraph end, calculated from document start
    html: Paragraph html
    text: Paragraph plain text i.e. segments plain texts separated by a space
    segments: list of Segments within the paragraph
    """
    start: int
    end: int
    html: str
    text: str
    segments: List[Segment]

    def __init__(self, start: int, end: int, html: str, text: str, segments: List[Segment]):
        self.start = start
        self.end = end
        self.html = html
        self.text = text
        self.segments = segments

    def __str__(self) -> str:
        if self.html:
            return self.html
        return super().__str__()


class Document:
    text: str
    paragraphs: List[Paragraph]

    def __init__(self, html: str):
        """
        Current document format stored in bigtable ccontains:
            spans for segments
            br(s) as paragraph separator
        :param html: document html
        """
        if html:
            soup = BeautifulSoup(html, 'html.parser')
            segment = None
            paragraph = None
            paragraph_segments = []
            paragraphs = []
            paragraph_start = 0
            paragraph_text = ''
            document_text = ''
            for element in soup.find_all():
                # update previous segment end position
                # because of invalid html tags `end = start + element_len` didn't work e.g. <br> should be <br/>
                if segment is not None:
                    segment.end = element.sourcepos
                    segment.html = html[segment.start:segment.end]
                    segment.start_text = segment.html.find(segment.text)
                    segment.end_text = segment.start_text + len(segment.text)
                    segment = None

                if paragraph is not None:
                    paragraph.end = element.sourcepos
                    paragraph.html = html[paragraph.start:paragraph.end]
                    paragraph = None

                if len(paragraph_segments) == 0:
                    paragraph_start = element.sourcepos

                tag = element.name
                if tag == 'span':
                    start = element.sourcepos
                    text = element.get_text().strip()
                    paragraph_text += text + ' '
                    attrs = element.attrs
                    segment = Segment(start, 0, tag, '', text, 0, 0, attrs)
                    paragraph_segments.append(segment)
                elif tag == 'br':
                    # double br
                    if element.nextSibling and element.nextSibling.name == 'br':
                        continue

                    if len(paragraph_segments) > 0:
                        paragraph = Paragraph(paragraph_start, 0, '', paragraph_text[0:-1], paragraph_segments)
                        paragraphs.append(paragraph)
                        document_text += paragraph_text
                        paragraph_segments = []
                        paragraph_text = ''
                else:
                    raise NotImplementedError()

            if segment is not None:
                segment.end = len(html)
                segment.html = html[segment.start:segment.end]

            if paragraph is not None:
                paragraph.end = len(html)
                paragraph.html = html[paragraph.start:paragraph.end]

            self.text = document_text[0:-1]
            self.paragraphs = paragraphs
        else:
            self.text = ''
            self.paragraphs = []

    def __str__(self) -> str:
        if self.text:
            return self.text
        return super().__str__()


class PlainDocument:
    text: str
    paragraphs: List[Paragraph]

    def __init__(self, plain_text: str, nlp):
        """
        Current extension document format stored in bigtable ccontains plain text and new lines
        WE DO DIFFERENT SEGMENTATION, SO COMPARING DOCUMENTS SEGMENTS WITH REST PAYLOAD SEGMENT WON'T WORK
        :param plain_text: document content
        """
        self.nlp = nlp
        if plain_text:
            paragraphs = []
            document_text = ''
            start = 0
            for match in re.finditer('(\\s)*(\\n)+(\\s)*(\\n)*', plain_text, re.MULTILINE):  # paragraph split
                end = match.start()  # text end, without the new lines separators
                text = plain_text[start:end]
                document_text += text + ' '

                end = match.end()  # with new lines separators
                html = plain_text[start:end]

                segments = self._extract_segments(text, start)
                paragraph = Paragraph(start, end, html, text, segments)
                paragraphs.append(paragraph)
                start = match.end()

            plain_text_len = len(plain_text)
            if plain_text_len > start:
                text = plain_text[start:plain_text_len]
                document_text += text + ' '
                segments = self._extract_segments(text, start)
                paragraph = Paragraph(start, plain_text_len, text, text, segments)
                paragraphs.append(paragraph)

            self.text = document_text[0:-1]
            self.paragraphs = paragraphs
        else:
            self.text = ''
            self.paragraphs = []

    def _extract_segments(self, paragraph_text, paragraph_start):
        segments = []
        for sent in self.nlp(paragraph_text, disable=['tagger', 'ner']).sents:
            segment = Segment(paragraph_start + sent.start_char, paragraph_start + sent.end_char, '', sent.text,
                              sent.text, sent.start_char, sent.end_char)
            segments.append(segment)
        return segments

    def __str__(self) -> str:
        if self.text:
            return self.text
        return super().__str__()



import os

from setuptools import setup, find_packages

here = os.path.abspath(os.path.dirname(__file__))

with open(os.path.join(here, "README.md"), encoding="utf-8") as f:
    long_description = f.read()


def python_version_bump_from_shell_file():
    with open(os.path.join(here, "VERSION"), encoding="utf-8") as f:
        __version__ = f.read().strip()
        # you can't trust the module to reach out of qai at runtime
        # so write it to a python file at build time
        # I haven't seen this before, but it is the best way to handle exposing version
        # in a way that works with python's module bs, but is accessible to other languages
        # like bash or groovy for CI/CD stuff
        with open(os.path.join(here, "qai", "version.py"), "w+", encoding="utf-8") as v:
            v.write("# Don't change this file, ../VERSION is the source of truth\n")
            v.write(f'__version__ = "{__version__}"')
    return __version__


extras = {
    "storage": ["google-cloud-bigtable==1.4.0", "beautifulsoup4==4.9.1"]
}

setup(
    name="qai",
    packages=find_packages(),
    author="Qordoba",
    author_email="sam.havens@qordoba.com",
    url="https://github.com/Qordobacode/library.qai.utilities",
    version=python_version_bump_from_shell_file(),
    license="unlicensed",
    long_description=long_description,
    long_description_content_type="text/markdown",
    python_requires=">=3.6.4",
    install_requires=["attrdict>=2.0.1", "sanic>=20.6.3", "spacy>=2.1.3"],
    extras_require=extras,
)
